// ---------------------------------------------------------------------------
//	FM Sound Generator
//	Copyright (C) cisc 1998, 2000.
// ---------------------------------------------------------------------------
//	$Id: fmgen.h,v 1.21 2000/02/18 02:05:20 cisc Exp $

#ifndef FM_GEN_H
#define FM_GEN_H

#include "types.h"

// ---------------------------------------------------------------------------
//	萔̂P
//	ÓIe[ũTCY

#define FM_LFOBITS		8
#define FM_TLBITS		7

// ---------------------------------------------------------------------------

#define FM_TLENTS		(1 << FM_TLBITS)
#define FM_LFOENTS		(1 << FM_LFOBITS)
#define FM_TLPOS		(FM_TLENTS/4)

// ---------------------------------------------------------------------------

namespace FM
{
	//	Types ----------------------------------------------------------------
	typedef int32	ISample;
	typedef int32 	Sample;

	enum { ARCSIZE = 74 };
	
	//	Tables (O[oȂ̂ asm QƂ̓) -----------------
	void MakeTable();
	void MakeTimeTable(uint ratio);
	extern uint32 tltable[];
	extern int32  cltable[];
	extern uint32 dltable[];
	extern int    pmtable[][FM_LFOENTS];
	extern uint   amtable[][FM_LFOENTS];
	extern uint32 lfotable[8];
	extern uint   lfoc;

	//	Operator -------------------------------------------------------------
	class Operator
	{
	public:
		Operator();
		static void MakeTable();
		static void	MakeTimeTable(uint ratio);
		
		ISample	Calc(ISample in);
		ISample	CalcL(ISample in);
		ISample CalcFB(uint fb);
		ISample CalcFBL(uint fb);
		void	Prepare();
		void	KeyOn();
		void	KeyOff();
		void	Reset();
		void	ResetFB();
		int		IsOn();
		
		void	SetDT(uint dt);
		void	SetMULTI(uint multi);
		void	SetTL(uint tl, bool csm);
		void	SetKS(uint ks);
		void	SetAR(uint ar);
		void	SetDR(uint dr);
		void	SetSR(uint sr);
		void	SetRR(uint rr);
		void	SetSL(uint sl);
		void	SetSSGEC(uint ssgec);
		void	SetFNum(uint fnum);
		void	SetDPBN(uint dp, uint bn);
		void	SetMode(bool modulator);
		void	SetAMON(bool on);
		void	SetMS(uint ms);
		void	Mute(bool);
		
		static void SetLFOC(uint c) { lfoc = c & (FM_LFOENTS - 1); }
	
	private:
		typedef uint32 Counter;
		
		ISample	out, out2;
		uint	dp;
		uint	bn;

	//	Phase Generator ------------------------------------------------------
		
		uint32	PGCalc();
		uint32	PGCalcL();

		uint	detune;		// Detune
		uint	multiple;	// Multiple
		uint32	pgcount;
		uint32	pgdcount;
		int32	pgdcountl;

	//	Envelope Generator ---------------------------------------------------
		
		enum	EGPhase { next, attack, decay, sustain, release, off };
		
		void	EGCalc();
		void	ShiftPhase(EGPhase nextphase);
		void	ShiftPhase2();
		void	SetEGRate(uint);
		void	EGUpdate();
		
		int		eglevel;	// EG ̏o͒l
		int		eglvnext;	//  phase Ɉڂl
		int32	egstep;		// EG ̎̕ψڂ܂ł̎
		int32	egstepd;	// egstep ̎ԍ
		int		egtransa;	// EG ω̊ (for attack)
		int		egtransd;	// EG ω̊ (for decay)
		int		egout;		// EG+TL 킹o͒l
		int		tlout;		// TL ̏o͒l
		
		uint	ksr;		// key scale rate
		EGPhase	phase;
		uint*	ams;
		int*	pms;
		uint8	ms;

		bool	key;		// current key state
		
		uint8	tl;			// Total Level	 (0-127)
		uint8	tll;		// Total Level Latch (for CSM mode)
		uint8	ar;			// Attack Rate   (0-63)
		uint8	dr;			// Decay Rate    (0-63)
		uint8	sr;			// Sustain Rate  (0-63)
		uint8	sl;			// Sustain Level (0-127)
		uint8	rr;			// Release Rate  (0-63)
		uint8	ks;			// Keyscale      (0-3)
		uint8	ssgtype;	// SSG-Type Envelop Control

		bool	amon;		// enable Amplitude Modulation
		bool	paramchanged;	// p[^XVꂽ
		bool	mute;
		
	//	Tables ---------------------------------------------------------------
		
		enum TableIndex { dldecay = 0, dlattack = 0x400, };
		
		static Counter ratetable[64];
		static uint32 multable[16];

		friend class Channel4;
		friend void __stdcall FM_NextPhase(Operator* op);
	};
	
	//	4-op Channel ----------------------------------------------------------
	class Channel4
	{
	public:
		Channel4() { SetAlgorithm(0); }
		
		ISample Calc();
		ISample CalcL();
		void SetFNum(uint fnum);
		void SetFB(uint fb);
		void SetAlgorithm(uint algo);
		int Prepare();
		void KeyControl(uint key);
		void Reset();
		void SetMS(uint ms);
		void Mute(bool);
		
		static void Calc2 (Channel4* ch, ISample* s0, ISample* s1);
		static void Calc2E(Channel4* ch, ISample* s0, ISample* s1);
		static void CalcL2(Channel4* ch, ISample* s0, ISample* s1);
		static void CalcL2E(Channel4* ch, ISample* s0, ISample* s1);

	private:
		static const uint8 fbtable[8];
		uint	fb;
		int		buf[4];
		int*	in[3];			// e OP ̓̓|C^
		int*	out[3];			// e OP ̏o̓|C^

	public:
		Operator op[4];
	};
}

#endif // FM_GEN_H
